// ==============================================================
//						   ProdConsModule
//
//						by Cristian Bontas
//
// ProdConsModule.cpp
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include "ProdConsModule.h"
#include <string.h>
#include <math.h>

// ==============================================================
// ProdConsModule class interface
// ==============================================================

ProdConsModule::ProdConsModule (OBJHANDLE hVessel, int flightmodel)
: VESSEL4 (hVessel, flightmodel)
{
}

ProdConsModule::~ProdConsModule ()
{
}

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void ProdConsModule::clbkSetClassCaps (FILEHANDLE cfg)
{
	char line[256];
	oapiReadItem_string(cfg, "ProdConsResCount", line);
	sscanf (line, "%d", &prodConsResCount);
	for(int i = 0; i < prodConsResCount; i++) {
		char abuf[256];
		sprintf(abuf,"ProdConsRes%d", (i + 1));
		oapiReadItem_string(cfg, abuf, line);
		sscanf (line, "%lf%lf%lf%lf%s", &production[i], &transfer[i], &capacity[i], &range[i], &prodConsResStr[i]);
	}
	for(int i = 0; i < prodConsResCount; i++) {
		PROPELLANT_HANDLE hpr = CreatePropellantResource (capacity[i]);
	}
	VESSEL4::clbkSetClassCaps(cfg);
}

void ProdConsModule::clbkPostCreation()
{
	VESSEL4::clbkPostCreation();
}

bool ProdConsModule::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
	// draw the default HUD
	VESSEL4::clbkDrawHUD (mode, hps, skp);

	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
//	sprintf(abuf,"%d", prodConsResCount);
//	skp->Text(20 / 1024 * hps->W, 160 / 720 * hps->H, abuf, strlen(abuf));
//	sprintf(abuf,"%d %d", hps->W, hps->H);
//	skp->Text(200 / 1280.0 * hps->W, 80 / 720.0 * hps->H, abuf, strlen(abuf));
	for(int idxRes = 0; idxRes < prodConsResCount; idxRes++) 
	{
		PROPELLANT_HANDLE ph = GetPropellantHandleByIndex(idxRes);
		double textColor = GetPropellantMass(ph) / (float)GetPropellantMaxMass(ph) * 511;
		if(textColor < 256) {
			textColor = 255 + (long)textColor * pow(16.0,2);
		} else {
			textColor = (511 - (long)textColor) + 255 * pow(16.0,2);
		}
		sprintf(debug, "%d", (long)textColor);
		skp->SetTextColor((long)textColor);
		char abuf[256];
		char timeRemaining[256];
		sprintf(timeRemaining, "999y");
		if(production[idxRes] < -1e-10) {
			double dTime = (oapiGetPropellantMass(ph) / (-production[idxRes])) / 365;
			double dYear = dTime;
			dTime = (dTime - (int)dTime) * 365;
			double dDay = dTime;
			dTime = (dTime - (int)dTime) * 24;
			double dHour = dTime;
			dTime = (dTime - (int)dTime) * 60;
			double dMin = dTime;
			dTime = (dTime - (int)dTime) * 60;
			double dSec = dTime;
			if(dYear > 1) {
				sprintf(timeRemaining, "%.0fy %.0fd", floor(dYear), floor(dDay));
			} else if(dDay > 1) {
				sprintf(timeRemaining, "%.0fd %.0fh", floor(dDay), floor(dHour));
			} else if(dHour > 1) {
				sprintf(timeRemaining, "%.0fh %.0fm", floor(dHour), floor(dMin));
			} else {
				sprintf(timeRemaining, "%.0fm %.0fs", floor(dMin), floor(dSec));
			}
		}
		sprintf(abuf,"%.2f %.2f %.2f %.2f %s %s", production[idxRes], transfer[idxRes], oapiGetPropellantMass(ph), range[idxRes], prodConsResStr[idxRes], timeRemaining);
		skp->Text((int)(190 / 1280.0 * hps->W), (int)((70 + 18 * idxRes) / 720.0 * hps->H), abuf, strlen(abuf));
	}
	return true;
}

void ProdConsModule::clbkPreStep (double simt, double simdt, double mjd)
{
	for(int idxRes = 0; idxRes < prodConsResCount; idxRes++) 
	{
//		totalTransfer[idxRes] = 0;
//		totalDiff[idxRes] = (prodConsRes[idxRes][0] - prodConsRes[idxRes][1]) / 3600 / 24 * simdt;
//		totalDiffForTransfer[idxRes] = abs(totalDiff[idxRes]);
		
		forTransfer[idxRes] = production[idxRes] / 24 / 3600 * simdt;
		PROPELLANT_HANDLE ph = GetPropellantHandleByIndex(idxRes);

		if(GetPropellantMass(ph) + forTransfer[idxRes] > GetPropellantMaxMass(ph)) {
			forTransfer[idxRes] -= GetPropellantMaxMass(ph) - GetPropellantMass(ph);
			SetPropellantMass(ph, GetPropellantMaxMass(ph));
		} else if(GetPropellantMass(ph) + forTransfer[idxRes] < 0) {
			forTransfer[idxRes] += GetPropellantMass(ph);
			SetPropellantMass(ph, 0);
		} else {
			SetPropellantMass(ph, GetPropellantMass(ph) + forTransfer[idxRes]);
			forTransfer[idxRes] = 0;
		}
		forTransfer[idxRes] += transfer[idxRes] / 3600 / 24 * simdt;

/*		if(totalDiff[idxRes] > -1e-10) {
			if (GetPropellantMass(ph) + totalDiff[idxRes] > GetPropellantMaxMass(ph)) {
				totalDiffForTransfer[idxRes] = GetPropellantMass(ph) + totalDiff[idxRes] - GetPropellantMaxMass(ph);
				SetPropellantMass(ph, GetPropellantMaxMass(ph));
			} else {
				totalDiffForTransfer[idxRes] = 0;
				SetPropellantMass(ph, GetPropellantMass(ph) + totalDiff[idxRes]);
			}
		} else {
			totalDiffForTransfer[idxRes] *= 2;
		}*/

		int noVessels = oapiGetVesselCount();
		for(int idxVessel = 0; idxVessel < noVessels; idxVessel++) {
			OBJHANDLE hVessel = oapiGetVesselByIndex(idxVessel);
			VESSEL* vessel = oapiGetVesselInterface(hVessel);
			if(strstr(vessel->GetClassNameA(),"ProdConsModule\\")) {
				if(abs(forTransfer[idxRes]) < 1e-10)
					break;
//				int pcm1ResNo = 0;
				ProdConsModule* pcm1 = dynamic_cast<ProdConsModule*>((VESSEL4*)vessel);
				VECTOR3 vec1, vec2;
				GetGlobalPos(vec1);
				pcm1->GetGlobalPos(vec2);
				if(!_strcmpi(pcm1->GetName(), GetName())) {
					continue;
				}
				if(getDistance(vec1,vec2) > range[idxRes]) {
					continue;
				}
				for(int idxResTarget = 0; idxResTarget < pcm1->getProdConsResCount(); idxResTarget++) {
					if(!_strcmpi(pcm1->getProdConsResStr(idxResTarget), getProdConsResStr(idxRes))) {
						PROPELLANT_HANDLE phTarget = pcm1->GetPropellantHandleByIndex(idxResTarget);
						if(GetPropellantMass(ph) + forTransfer[idxRes] > GetPropellantMaxMass(ph)) {
							forTransfer[idxRes] = GetPropellantMaxMass(ph) - GetPropellantMass(ph);
						} else if(GetPropellantMass(ph) + forTransfer[idxRes] < 0) {
							forTransfer[idxRes] = -GetPropellantMass(ph);
						}
						if(pcm1->GetPropellantMass(phTarget) - forTransfer[idxRes] > pcm1->GetPropellantMaxMass(phTarget)) {
							SetPropellantMass(ph, GetPropellantMass(ph) - (pcm1->GetPropellantMaxMass(phTarget) - pcm1->GetPropellantMass(phTarget)));
							forTransfer[idxRes] += pcm1->GetPropellantMaxMass(phTarget) - pcm1->GetPropellantMass(phTarget);
							pcm1->SetPropellantMass(phTarget, pcm1->GetPropellantMaxMass(phTarget));
						} else if(pcm1->GetPropellantMass(phTarget) - forTransfer[idxRes] < 0) {
							SetPropellantMass(ph, GetPropellantMass(ph) + pcm1->GetPropellantMass(phTarget));
							forTransfer[idxRes] -= pcm1->GetPropellantMass(phTarget);
							pcm1->SetPropellantMass(phTarget, 0);
						} else {
							SetPropellantMass(ph, GetPropellantMass(ph) + forTransfer[idxRes]);
							pcm1->SetPropellantMass(phTarget, pcm1->GetPropellantMass(phTarget) - forTransfer[idxRes]);
							forTransfer[idxRes] = 0;
						}
/*						if(forTransfer[idxRes] < 1e-10
							&& (pcm1->getProdConsRes(idxResTarget,0) - pcm1->getProdConsRes(idxResTarget,1)) >= -0.0001) {
								if(totalDiffForTransfer[idxRes] < pcm1->GetPropellantMass(phTarget)) {
									if(GetPropellantMass(ph) + totalDiffForTransfer[idxRes] <= GetPropellantMaxMass(ph)) {
										totalTransfer[idxRes] += -totalDiffForTransfer[idxRes];
										pcm1->SetPropellantMass(phTarget, pcm1->GetPropellantMass(phTarget) - totalDiffForTransfer[idxRes]);
										SetPropellantMass(ph, GetPropellantMass(ph) + totalDiffForTransfer[idxRes]);
										totalDiffForTransfer[idxRes] = 0;
									} else {
										totalTransfer[idxRes] += -(GetPropellantMaxMass(ph) - GetPropellantMass(ph));
										pcm1->SetPropellantMass(phTarget, pcm1->GetPropellantMass(phTarget) - (GetPropellantMaxMass(ph) - GetPropellantMass(ph)));
										SetPropellantMass(ph, GetPropellantMass(ph) + (GetPropellantMaxMass(ph) - GetPropellantMass(ph)));
										totalDiffForTransfer[idxRes] += GetPropellantMaxMass(ph) - GetPropellantMass(ph);
									}
								} else {
									if(GetPropellantMass(ph) + pcm1->GetPropellantMass(phTarget) <= GetPropellantMaxMass(ph)) {
										totalTransfer[idxRes] += -pcm1->GetPropellantMass(phTarget);
										pcm1->SetPropellantMass(phTarget, 0);
										SetPropellantMass(ph, GetPropellantMass(ph) + pcm1->GetPropellantMass(phTarget));
										totalDiffForTransfer[idxRes] -= pcm1->GetPropellantMass(phTarget);
									} else {
										totalTransfer[idxRes] += -(GetPropellantMaxMass(ph) - GetPropellantMass(ph));
										pcm1->SetPropellantMass(phTarget, pcm1->GetPropellantMass(phTarget) - (GetPropellantMaxMass(ph) - GetPropellantMass(ph)));
										SetPropellantMass(ph, GetPropellantMass(ph) + (GetPropellantMaxMass(ph) - GetPropellantMass(ph)));
										totalDiffForTransfer[idxRes] += GetPropellantMaxMass(ph) - GetPropellantMass(ph);
									}
								}
						}*/
/*						if(totalDiff[idxRes] > -1e-10
							&& totalDiffForTransfer[idxRes] > 0) {
								if(pcm1->GetPropellantMass(phTarget) + totalDiffForTransfer[idxRes] < pcm1->GetPropellantMaxMass(phTarget)) {
									totalTransfer[idxRes] += totalDiffForTransfer[idxRes];
									pcm1->SetPropellantMass(phTarget, pcm1->GetPropellantMass(phTarget) + totalDiffForTransfer[idxRes]);
									totalDiffForTransfer[idxRes] = 0;
								} else {
									totalTransfer[idxRes] += pcm1->GetPropellantMaxMass(phTarget) - pcm1->GetPropellantMass(phTarget);
									pcm1->SetPropellantMass(phTarget, pcm1->GetPropellantMaxMass(phTarget));
									totalDiffForTransfer[idxRes] -= pcm1->GetPropellantMaxMass(phTarget) - pcm1->GetPropellantMass(phTarget);
								}
						}*/
						break;
					}
				}
			}
		}
/*		if(totalDiff[idxRes] < 1e-10) {
			if(GetPropellantMass(ph) + totalDiff[idxRes] > 0) {
				SetPropellantMass(ph,GetPropellantMass(ph) + totalDiff[idxRes]);
			} else {
				SetPropellantMass(ph, 0);
			}
		}*/
	}
}

/*int ProdConsModule::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {

	if (!down) return 0;

	if (KEYMOD_SHIFT (kstate)) {	 
		switch (key) {
		} 
	} else if (KEYMOD_CONTROL (kstate)) { 
		switch (key) {
		} 
	} else {
		switch (key) {
			case OAPI_KEY_G: // 

		}
	}
	return 1;
}*/

double ProdConsModule::getDistance(VECTOR3 vec1, VECTOR3 vec2) {
	return sqrt(pow(vec1.x - vec2.x,2) + pow(vec1.y - vec2.y,2) + pow(vec1.y - vec2.y,2));
}

double ProdConsModule::getProduction(int idx) {
	return production[idx];
}

double ProdConsModule::getTransfer(int idx) {
	return transfer[idx];
}

double ProdConsModule::getCapacity(int idx) {
	return capacity[idx];
}

double ProdConsModule::getRange(int idx) {
	return range[idx];
}

/*double ProdConsModule::getProdConsRes(int x, int y) {
	return prodConsRes[x][y];
}*/

char* ProdConsModule::getProdConsResStr(int x) {
	return prodConsResStr[x];
}

int ProdConsModule::getProdConsResCount() {
	return prodConsResCount;
}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new ProdConsModule (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (ProdConsModule*)vessel;
}
